/******************************************************************/

#include "system.h"
#include "filosofo.h"

/******************************************************************/

static int NroFilosofo;
static ProcesoType*Proceso;

/******************************************************************/
void FilosofoSignal(int sig)
{
    signal(SIG_COMIENZAPENSAR,(SIG_PF)FilosofoSignal);
    signal(SIG_RECIVEPALITOS,(SIG_PF)FilosofoSignal);
    if (DatosGlobales->TerminarFlag)
    {
	DettachSharedMem();
        LogPrint("(Filosofo %i) Terminado.",NroFilosofo);
	exit(0); 
    }	

    if (sig==SIG_COMIENZAPENSAR)
    {
        Assert(Proceso->Estado==ESTADO_PENSANDO);
        //LogPrint("(Filosofo %i) Signal SIG_COMIENZA_PENSAR",NroFilosofo);
    }
    else if (sig==SIG_RECIVEPALITOS)
    {
        Assert(Proceso->Estado==ESTADO_COMIENDO);
        //LogPrint("(Filosofo %i) Signal SIG_RECIVEPALITOS",NroFilosofo);
    }
    else
        Assert(FALSE);
}

/******************************************************************/
void Pensar()
{
    if(Proceso->Estado!=ESTADO_PENSANDO)
        pause();

    long int CPuUsadaAlComienzoPensar=GetCPUTime();
    LogPrint("(Filosofo %i) Comienza Pensar",NroFilosofo);

    Proceso->TiempoPensado=0;

    int TiempoAPensar=Proceso->TiempoPensar;
    Assert(Proceso->Estado==ESTADO_PENSANDO);

    while(Proceso->TiempoPensado < TiempoAPensar)
    {
        long int CPuUsada=GetCPUTime();
        Proceso->TiempoPensado=CPuUsada-CPuUsadaAlComienzoPensar;
    };

    LogPrint("(Filosofo %i) Fin Pensar",NroFilosofo);

    //Indico que necesito los palitos
    Proceso->Estado=ESTADO_ESPERANDO_PALITOS;
    SendSignal(DatosGlobales->pid_AdministradorPalitos,SIG_DARPALITOS);
    SendSignal(DatosGlobales->pid_Planificador,SIG_FINPENSAR);
}

/******************************************************************/
void Comer()
{
    //Espero que me avisen que tengo los palitos
    if(Proceso->Estado!=ESTADO_COMIENDO)
        pause();
    LogPrint("(Filosofo %i) Comienzo Comer",NroFilosofo);

    Assert(Proceso->Estado==ESTADO_COMIENDO);
    sleep(Proceso->TiempoComer);

    LogPrint("(Filosofo %i) Fin Comer",NroFilosofo);
    Proceso->Estado=ESTADO_OCIOSO;
    SendSignal(DatosGlobales->pid_AdministradorPalitos,SIG_DARPALITOS);
}

/******************************************************************/
void Filosofo(int NroFil)
{
    NroFilosofo=NroFil;
    LogPrint("(Filosofo %i) Comienzo Inicializacion",NroFilosofo);

    AttachSharedMem();

    signal(SIG_COMIENZAPENSAR,(SIG_PF)FilosofoSignal);
    signal(SIG_RECIVEPALITOS,(SIG_PF)FilosofoSignal);
    Assert(DatosGlobales->Check==12345);
    Proceso=DatosGlobales->Proceso + NroFil;

    Proceso->pid=getpid();
    Proceso->TiempoPensado=0;
    Proceso->ComienzoDelPeriodo=GetHoraActual();

    LogPrint("(Filosofo %i) Fin Inicializacion",NroFilosofo);
    Proceso->Estado=ESTADO_ESPERANDO_PENSAR;

    nice(1);

    while(!DatosGlobales->TerminarFlag)
    {
        Pensar();
        Comer();
    }

    DettachSharedMem();
    LogPrint("(Filosofo %i) Termina",NroFilosofo);
}

/******************************************************************/
/******************************************************************/

