/******************************************************************/

#include "system.h"
#include "filosofo.h"

/******************************************************************/

extern VoidF Filosofo;
extern VoidF Planificador;
extern VoidF AdministradorPalitos;
extern VoidF Supervisor;
extern VoidF Aperiodicas;

long int HoraInicioDelPrograma;

/******************************************************************/
void LogPrint(const char *fmt, ... )
{
const int size=1000;
va_list marker;
char buffer[size];

    va_start( marker, fmt );
    vsprintf(buffer,fmt,marker);
    va_end( marker );

    long Segundos=GetHoraActual()-HoraInicioDelPrograma;
    fprintf(stdout,"%li: %s\n",Segundos,buffer);
	fflush(stdout);
}

/******************************************************************/
long GetHoraActual()
{
#ifndef __BORLANDC__
struct timeval T;
struct timezone tm;

    gettimeofday(&T,&tm);
    return(T.tv_sec);
#else
	return(1);
#endif
}

/******************************************************************/
long GetCPUTime()
//Devuelve el tiempo de procesador usado por el proceso en segundos
{
#ifndef __BORLANDC__
struct tms T;
    times(&T);

    //QWERTY Averiguar cual de los returns es el correcto.
    //La idea es tomar el tiempo que uso la CPU el proceso sin importar
    //si en user o kernel mode. Sin embargo, no se si tms_stime incluye el
    //tiempo que el proceso esta dormido o no. Creo que no lo incluye.

    //return( T.tms_utime/HZ);
    return( (T.tms_utime+T.tms_stime)/HZ);

#else
    return(1);
#endif
}

/******************************************************************/
void SendSignal(int pid,SIG_ENUM Sig)
{
    Assert(pid!=0);

    kill(pid,(int)Sig);
}

/******************************************************************/
int Fork(VoidF f,int dato)
{
int id;
    id=fork();
    Assert(id!=-1);

    //termino la funcion para el padre
    if (id!=0)
        return(id);

    //Llamo a la funcion que es el punto de entrada
    f(dato);
    exit(0);
    return(0);
}

/******************************************************************/
void GetListaDeNumeros(FILE* fp,int* Lista,int MaxCantEnLista,int CantEnLista)
{
char buffer[1000];
int Auxiliar;
	Assert(LeerRenglonDeArchivoSinComentario(buffer,sizeof(buffer),fp,Auxiliar)!=TRUE);

	ConvertirAVector(buffer,Lista,Auxiliar,MaxCantEnLista);
	Assert(Auxiliar==CantEnLista);
}

/******************************************************************/
void InicializarMemCompartida()
{
ProcesoType*Proceso;
int i;

int Lista[50];
FILE* fp=fopen("datos.txt","rt");
	if (fp==NULL)
		Error(FatalError,"No se encuentra el archivo 'datos.txt'");

	GetListaDeNumeros(fp,Lista,50,1);
	Assert(Lista[0]>=0);	Assert(Lista[0]<=2);
	ALGORITMO_DE_PLANIFICACION=(ALGOR_DE_PLANIF_ENUM)Lista[0];

	GetListaDeNumeros(fp,Lista,50,2);
	Assert(Lista[0]>1);	Assert(Lista[0]<Lista[1]);
	MinTiempoMozo=Lista[0];
	MaxTiempoMozo=Lista[1];

	CreateSharedMem();
	AttachSharedMem();

	DatosGlobales->Check=12345;
	DatosGlobales->TerminarFlag=FALSE;
	DatosGlobales->pid_Planificador         =0;
	DatosGlobales->pid_Supervisor           =0;
	DatosGlobales->pid_Aperiodicas          =0;
	DatosGlobales->pid_AdministradorPalitos =0;

	DatosGlobales->CantFilosofosMuertos=0;
	DatosGlobales->CantMozosMuertos=0;

	for(i=0;i<CANT_FIL+1;i++)
	{
		GetListaDeNumeros(fp,Lista,50,4);
		Proceso=DatosGlobales->Proceso + i;

		Proceso->pid=0;
		Proceso->Periodo      =Lista[0];
		Proceso->Meta         =Lista[1];
		Proceso->TiempoPensar =Lista[2];
		Proceso->TiempoComer  =Lista[3];

		Proceso->TiempoPensado=0;
		Proceso->ComienzoDelPeriodo=GetHoraActual();
		Proceso->Estado=ESTADO_NO_CREADO;

		//Esto es para el mozo. Lo hago para todos asi la memoria empieza bien inicializada
		Proceso->Limpiando=-1;
		if (i==PMOZO)
			Proceso->Periodo=Proceso->Meta;

	}
	fclose(fp);
}

/******************************************************************/
void Finalizar()
{
	LogPrint("(MAIN) Mandando mensaje de finalizacion...");
	DatosGlobales->TerminarFlag=TRUE;

	//Mando mensajes para que todos salgan de la pausa y chequeen el archivo
	SendSignal(DatosGlobales->pid_Planificador,SIG_TERMINAR);
	SendSignal(DatosGlobales->pid_Supervisor,SIG_TERMINAR);
	SendSignal(DatosGlobales->pid_AdministradorPalitos,SIG_TERMINAR);
	SendSignal(DatosGlobales->pid_Aperiodicas,SIG_TERMINAR);

	for(int i=0;i<CANT_FIL+1;i++)
	{
		ProcesoType*Proceso=DatosGlobales->Proceso + i;
		if (Proceso->Estado==ESTADO_NO_CREADO)
			continue;
		Proceso->Estado=ESTADO_PENSANDO;
		SendSignal(Proceso->pid,SIG_TERMINAR);
	}
    int fm = DatosGlobales->CantFilosofosMuertos;
    int mm = DatosGlobales->CantMozosMuertos;
    LogPrint("FILOSOFOS MUERTOS: %i",fm);
    LogPrint("MOZOS MUERTOS: %i",mm);


    //Espero 5segundos para dar tiempo a que todos terminen
    sleep(5);
    DettachSharedMem();
    DestroySharedMem();
}

/******************************************************************/
void EsperarFilosofos()
{
int todosCreados;
	do{
	todosCreados=TRUE;
	for (int i=0;i<CANT_FIL;i++)
	{
			if (DatosGlobales->Proceso[i].Estado==ESTADO_NO_CREADO)
			todosCreados=FALSE;
	}
	}while(!todosCreados);
}

/******************************************************************/
int main(int argc,char**argv)
{
	(void)argc;(void)argv;

	printf("\Filosofos - Version 1.0 - Bs. As. Argentina\n");
	printf("\tGaligniana, Daniela\n");
	printf("\tCanton, Gabriel\n");
	printf("\tGrippo, Daniel\n");
    printf("\tRobak, Dario\n\n\n");


    InicializarMemCompartida();
    
    //Espero el comienzo de un segundo
    HoraInicioDelPrograma=GetHoraActual();
    while(GetHoraActual()==HoraInicioDelPrograma);
    HoraInicioDelPrograma=GetHoraActual();
    	
    LogPrint("(MAIN) Programa Comienza");

    //LogPrint("(MAIN) Comienzan los Fork de filosofos");

	Fork(Filosofo,0);
	Fork(Filosofo,1);
	Fork(Filosofo,2);
	Fork(Filosofo,3);
    Fork(Filosofo,4);

    //Esperar que se inicialicen todos los filosofos
    EsperarFilosofos();		
 
    //LogPrint("(MAIN) Comienzan los Fork de tareas de planificacion");
    Fork(Planificador);
    Fork(AdministradorPalitos);
    Fork(Supervisor);
    Fork(Aperiodicas);

    LogPrint("(MAIN) Fin de los Fork()");

    //Espero que se toque una tecla
    getchar();

    Finalizar();
    LogPrint("(MAIN) Termina");
    return(0);
}

/******************************************************************/
/******************************************************************/
/******************************************************************/
/******************************************************************/
/******************************************************************/
/******************************************************************/

