#include "system.h"
#include "filosofo.h"

/******************************************************************/

    //Esta varible guarda el filosofo (o mozo) que esta pensando. -1 indica
    //que nadie esta pensando
    int FilActual=-1;

/******************************************************************/

	ALGOR_DE_PLANIF_ENUM ALGORITMO_DE_PLANIFICACION=TAZA_MONOTONICA;

/******************************************************************/
int CalcularTiempoDeAlarma()
{
long int Tiempo=999999999;
int HayFilCreadoFlag=FALSE;
ProcesoType*Proceso;

	if(ALGORITMO_DE_PLANIFICACION==MENOR_FLEXIBILIDAD_PRIMERO)
		return(1);

    //Calculo el fin de periodo mas cercano
    for (int i=0;i<CANT_FIL;i++)
    {
        Proceso=DatosGlobales->Proceso + i;

        if( Proceso->Estado==ESTADO_NO_CREADO)
            continue;

        HayFilCreadoFlag=TRUE;
        long int
		t=Proceso->Periodo-(GetHoraActual()-Proceso->ComienzoDelPeriodo);
        if (t<=0)
            return(1);

        Tiempo=MIN(Tiempo,t);
    }

    //Si no hay ningun filosofo creado, retorno 1segundo
    if (!HayFilCreadoFlag)
        return(1);

    return(Tiempo);
}


/******************************************************************/
long Prioridad(int Filosofo)
//Se da el procesador a la menor prioridad
{
    switch(ALGORITMO_DE_PLANIFICACION)
    {
        case TAZA_MONOTONICA:
            return DatosGlobales->Proceso[Filosofo].Periodo;

        case META_ANTERIOR_PRIMERO:
            return
                DatosGlobales->Proceso[Filosofo].ComienzoDelPeriodo+
                DatosGlobales->Proceso[Filosofo].Meta;

        case MENOR_FLEXIBILIDAD_PRIMERO:
        {
            long FaltaPensar=
                DatosGlobales->Proceso[Filosofo].TiempoPensar-
                DatosGlobales->Proceso[Filosofo].TiempoPensado;

            return
                DatosGlobales->Proceso[Filosofo].Meta-FaltaPensar;
        }
    }

	Assert(0);
    return 0;
}

/******************************************************************/
int ElegirFilosofo()
{
int FilosofoAPensar=-1;
long PrioridadMenor=-1;
ProcesoType*Proceso;

    for (int i=0;i<CANT_FIL+1;i++)  //+1 para el mozo
	{
        Proceso=DatosGlobales->Proceso + i;
        if( (Proceso->Estado!=ESTADO_ESPERANDO_PENSAR) &&
            (Proceso->Estado!=ESTADO_PENSANDO)
          )
            continue;

        if( (FilosofoAPensar==-1) ||
            (Prioridad(i) < PrioridadMenor)
          )
        {
            FilosofoAPensar=i;
            PrioridadMenor=Prioridad(i);
        }
    }
    return FilosofoAPensar;
}

/******************************************************************/
void Planificar()
{
ProcesoType*Proceso;

    //Me aseguro que FilActual sea valido. Se invalida cuando un proceso termina
    if( (FilActual!=-1) &&
        (DatosGlobales->Proceso[FilActual].Estado!=ESTADO_PENSANDO)
      )
        FilActual=-1;


    //Veo si hay que actualizar algun periodo
    for (int i=0;i<CANT_FIL;i++)
    {
        Proceso=DatosGlobales->Proceso + i;
        if(Proceso->Estado!=ESTADO_OCIOSO)
			continue;
        if(GetHoraActual()>=Proceso->ComienzoDelPeriodo+Proceso->Periodo)
        {
            LogPrint("(Planificador) Inicializo Periodo de filosofo %i",i);

            Proceso->Estado=ESTADO_ESPERANDO_PENSAR;
            Proceso->ComienzoDelPeriodo+=Proceso->Periodo;
	    	SendSignal(DatosGlobales->pid_Supervisor,SIG_NUEVOPROCESO);
        }
    }

    int FilosofoAPensar=ElegirFilosofo();

    if(FilosofoAPensar==-1)
    {
        LogPrint("(Planificador) Ningun filosofo para pensar!");
        FilActual=-1;

        //me despierto en 1segundo para ver si hay algun filosofo que
        //despertar y ponerlo a pensar. Igual no hay otra cosa para
        //hacer con la CPU...
        alarm(1);
        return;
    }

    //Para al filosofo que esta pensando y hace continuar al elegido
    if (FilActual!=FilosofoAPensar)
    {
        if (FilActual!=-1)
        {
            LogPrint("(Planificador) Para al filosofo %i",FilActual);
            SendSignal(DatosGlobales->Proceso[FilActual].pid,SIG_PARAR);
        }

        FilActual=FilosofoAPensar;

	if (FilActual != 5) 
             LogPrint("(Planificador) Pone a pensar al filosofo %i",FilActual);
	else LogPrint ("(Planificador) Pone a pensar al MOZO");

        if (DatosGlobales->Proceso[FilActual].Estado==ESTADO_PENSANDO)
            SendSignal(DatosGlobales->Proceso[FilActual].pid,SIG_CONTINUAR);
        else if (DatosGlobales->Proceso[FilActual].Estado==ESTADO_ESPERANDO_PENSAR)
        {
            DatosGlobales->Proceso[FilActual].Estado=ESTADO_PENSANDO;
            SendSignal(DatosGlobales->Proceso[FilActual].pid,SIG_COMIENZAPENSAR);
        }
		else
			Assert(1==2);
	}

	FilActual=FilosofoAPensar;
	alarm(CalcularTiempoDeAlarma());
}

/******************************************************************/
void PlanificadorSignal(int sig)
{
	signal(SIG_FINPENSAR,(SIG_PF)PlanificadorSignal);
	signal(SIG_REPLANIFICAR,(SIG_PF)PlanificadorSignal);
	signal(SIGALRM,(SIG_PF)PlanificadorSignal);

//	LogPrint("(Planificador) Alarma :%i",sig);
}

/******************************************************************/
void Planificador(int nada)
{
	(void)nada;
	LogPrint("(Planificador) Comienzo Inicializacion");

	AttachSharedMem();
	DatosGlobales->pid_Planificador=getpid();

    //Inicializo el comienzo de los periodos de los filosofos
    for (int i=0;i<CANT_FIL;i++)
	DatosGlobales->Proceso[i].ComienzoDelPeriodo=GetHoraActual();
	
    signal(SIG_FINPENSAR,(SIG_PF)PlanificadorSignal);
    signal(SIG_REPLANIFICAR,(SIG_PF)PlanificadorSignal);
    signal(SIGALRM,(SIG_PF)PlanificadorSignal);

	LogPrint("(Planificador) Fin Inicializacion");

    while(!DatosGlobales->TerminarFlag)
    {
		WaitSemaf();
		Planificar();
		SignalSemaf();

		pause();
    }
    DettachSharedMem();
    LogPrint("(Planificador) Termina");
}


/******************************************************************/
/******************************************************************/
