#include 	<sys/types.h>
#include 	<sys/ipc.h>
#include 	<sys/sem.h>
#include 	<stdio.h>
#include 	<stdlib.h>

#include 	<errno.h>
extern int 	errno;

#define BIGCOUNT 	10000		/* Valor inicial del contador de procesos */


static struct sembuf	op_lock[2] = {
		2, 0, 0,		/* wait for [2] (lock) to equal 0 */
		2, 1, SEM_UNDO  /* then increment [2] to 1 - this lock it */
						/* UNDO to release the lock if processes exits
						   before explicitly unlocking */
};

static struct sembuf	op_endcreate[2] = {
		1, -1, SEM_UNDO,/* decrement [1] (proc counter) with undo on exit */
						/* UNDO to adjust proc counter if process exits
						   before explicitly calling sem_close() */
		2, -1, SEM_UNDO /* then decrement [2] (lock) back to 0 */
};

static struct sembuf	op_open[1] = {
		1, -1, SEM_UNDO /* decrement [1] (proc counter) with undo on exits */
};

static struct sembuf	op_close[3] = {
		2, 0, 0,		/* wait for [2] (lock) to equal 0 */
		2, 1, SEM_UNDO, /* then increment [2] to 1 - this lock it */
		1, 1, SEM_UNDO	/* then increment [1] (proc counter) */
};

static struct sembuf	op_unlock[1] = {
		2, -1, SEM_UNDO /* decrement [2] (lock) back to 0 */
};

static struct sembuf 	op_op[1] = {
		0, 99, SEM_UNDO /* decrement or increment [0] with undo on exit */
						/* the 99 is set to the actual amount to add
						   or subtract (positive or negative) */
};

//union {
//		int				val;
//		struct semid_ds *buf;
//		ushort			*array;
//} semnum;
semun semctl_arg;

void err_sys(char*p)
{
	printf("Error:%s\n",p);
	exit(-1);
}

void err_dump(char*p)
{err_sys(p);}

/* Crear un semaforo con un valor inicial especifico */

int sem_create(int key, int initval)
{
	register int	id, semval;

	if (key == IPC_PRIVATE)
		return(-1);		/* not intended for private semaphore */
	else if (key == (key_t) -1)
		return(-1);		/* probably an ftok() error by caller */

again:
	if ( (id = semget(key, 3, 0666 | IPC_CREAT)) < 0)
		return(-1);		/* permission problem or tables full */

	if (semop(id, &op_lock[0], 2) < 0) {
		if (errno == EINVAL)
			goto again;
		err_sys("can't lock");
	}

	semctl_arg.val = 0;
	if ( (semval = semctl(id, 1, GETVAL,semctl_arg)) < 0)
		err_sys("can't GETVAL");

	if (semval == 0) {
		semctl_arg.val = initval;
		if (semctl(id, 0, SETVAL, semctl_arg) < 0)
			err_sys("can't SETVAL[0]");

		semctl_arg.val = BIGCOUNT;
		if (semctl(id, 1, SETVAL, semctl_arg) < 0)
			err_sys("can't SETVAL[1]");
	}

	if (semop(id, &op_endcreate[0], 2) < 0)
		err_sys("can't end create");
	return(id);
}

/* Abrir un semaforo que debe existir */

int sem_open(int key)
{
	register int	id;

	if (key == IPC_PRIVATE)
		return(-1);
	else if (key == (key_t) -1)
		return(-1);

	if ( (id = semget(key, 3, 0)) < 0)
		return(-1);

	if (semop(id, &op_open[0], 1) < 0)
		err_sys("can't open");
	return(id);
}


/* Borrar un semaforo */

sem_rm(int id)
{
	semctl_arg.val = 0;
	if (semctl(id, 0, IPC_RMID, semctl_arg) < 0)
		err_sys("can't IPC_RMID");
}


/* Cerrar un semaforo */
sem_close(int id)
{
	register int	semval;

	if (semop(id, &op_close[0], 3) < 0)
		err_sys("can't semop");

	semctl_arg.val = 0;
	if ( (semval = semctl(id, 1, GETVAL, semctl_arg)) < 0)
		err_sys("can't GETVAL");

	if (semval > BIGCOUNT)
		err_dump("sem[1] > BIGCOUNT");
	else if (semval == BIGCOUNT)
		sem_rm(id);
	else
		if (semop(id, &op_unlock[0], 1) < 0)
			err_sys("can't unlock");
}

void sem_op(int id,int val)
{
	if ((op_op[0].sem_op=val)==0)
		err_sys(" can't have val==0" );
	if (semop(id,&op_op[0],1)<0)
		err_sys(" semop error" );
	
}

int sem_wait(int id)
{
	sem_op(id,-1);
}

int sem_signal(int id)
{
	sem_op(id,1);
}


