#include "system.h"
#include "filosofo.h"

extern VoidF Filosofo;

/******************************************************************/
void MatarFilosofo(int i)
{
ProcesoType*Proceso=DatosGlobales->Proceso + i;

    if (GetHoraActual() >= Proceso->ComienzoDelPeriodo+Proceso->Meta )
    {
        LogPrint("(Supervisor) Se mata al filosofo %i",i);
        Proceso->Estado=ESTADO_NO_CREADO;
        SendSignal(Proceso->pid,SIG_MATAR);
        DatosGlobales->CantFilosofosMuertos++;

        Fork(Filosofo,i);
        SendSignal(DatosGlobales->pid_Planificador,SIG_REPLANIFICAR);
    }
}

/******************************************************************/
void SetearAlarma(void)
{
ProcesoType*Proceso;
long int Tiempo=9999999999;
    //Calculo la meta mas cercana

    //Veo si hay un filosofo no creado.
    // Si es asi, pongo una alarma de 1segundo para esperar a que se cree
    // y asi saber cual es su meta.
    for (int i=0;i<CANT_FIL;i++)
    {
        if(DatosGlobales->Proceso[i].Estado==ESTADO_NO_CREADO)
        {
            alarm(1);
            return;
        }
    }

    if (DatosGlobales->Proceso[PMOZO].Estado!=ESTADO_NO_CREADO)
    {
        Proceso=DatosGlobales->Proceso + PMOZO;
        Tiempo=Proceso->ComienzoDelPeriodo+Proceso->Meta-GetHoraActual();
        if (Tiempo<=0)
        {
	    alarm(1);
            return;
	}
    }

    for (i=0;i<CANT_FIL;i++)
    {
        Proceso=DatosGlobales->Proceso + i;


     if( (Proceso->Estado!=ESTADO_ESPERANDO_PENSAR) &&
         (Proceso->Estado!=ESTADO_PENSANDO) &&
	 (Proceso->Estado!=ESTADO_ESPERANDO_PALITOS)
       )
          continue;

        long int t=Proceso->ComienzoDelPeriodo+Proceso->Meta-GetHoraActual();
        if (t<=0)
        {
            MatarFilosofo(i);
            alarm(1);  //Espero 1 segundo para que se inicialie el filosofo
                       //y pueda saber que meta tiene
            return;
        }

        Tiempo=MIN(Tiempo,t);
    }

    Assert(Tiempo>0);
    alarm(Tiempo);
}

/******************************************************************/
void Supervisar(void)
{
ProcesoType*Proceso;
    //Ver que no se tenga que matar al mozo
    Proceso=DatosGlobales->Proceso + PMOZO;
    if( (Proceso->Estado!=ESTADO_NO_CREADO) &&
        (Proceso->Estado!=ESTADO_OCIOSO) &&
	(Proceso->Limpiando != CANT_FIL-1 
	|| Proceso->Estado != ESTADO_COMIENDO))
    {
        if (GetHoraActual() >= Proceso->ComienzoDelPeriodo+Proceso->Meta )
        {
            LogPrint("(Supervisor) Se mata al mozo");
            DatosGlobales->Proceso[PMOZO].Estado=ESTADO_NO_CREADO;
            SendSignal(DatosGlobales->Proceso[PMOZO].pid,SIG_MATAR);
    	    SendSignal (DatosGlobales->pid_Planificador, SIG_REPLANIFICAR);
            DatosGlobales->CantMozosMuertos++;
        }
    }

    //Ver que no halla que matar a ningun filosofo
    for (int i=0;i<CANT_FIL;i++)
    {
        Proceso=DatosGlobales->Proceso + i;
        //Veo si hay que matar
        if( (Proceso->Estado!=ESTADO_ESPERANDO_PENSAR) &&
            (Proceso->Estado!=ESTADO_PENSANDO) &&
            (Proceso->Estado!=ESTADO_ESPERANDO_PALITOS)
          )
          continue;

        MatarFilosofo(i);
    }
}

/******************************************************************/
void SupervisorSignal(int sig)
{
    signal(SIGALRM,(SIG_PF)SupervisorSignal);
    signal(SIG_NUEVOPROCESO,(SIG_PF)SupervisorSignal);

    //LogPrint("(Supervisor) Alarma: %i",sig);
}

/******************************************************************/
void Supervisor(int nada)
{
    (void)nada;
    LogPrint("(Supervisor) Comienzo Inicializacion");

    AttachSharedMem();
    DatosGlobales->pid_Supervisor=getpid();
    signal(SIG_NUEVOPROCESO,(SIG_PF)SupervisorSignal);
    signal(SIGALRM,(SIG_PF)SupervisorSignal);

    DatosGlobales->CantFilosofosMuertos=0;
    DatosGlobales->CantMozosMuertos=0;

    LogPrint("(Supervisor) Fin Inicializacion");

    SetearAlarma();
    pause();

    while(!DatosGlobales->TerminarFlag)
	{
		WaitSemaf();

		Supervisar();
		SetearAlarma();

		SignalSemaf();
		pause();

    }

    DettachSharedMem();
    LogPrint("(Supervisor) Termina");
}


